// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.
//
// Class to evaluate translation quality. Currently supports BLEU and a few smoothed sentence-level versions of BLEU.

#ifndef MTEVAL_H_
#define MTEVAL_H_

#include <cstdlib>
#include <map>
#include <math.h>
#include "utils.h"

static const double COUNT_SMOOTH_VALUE = 0.01;

enum Metric {STANDARD_BLEU, LIN_OCH_BLEU_PLUS_ONE, BLEU_PLUS_ONE_ALL_N, BLEU_PLUS_ONE_ALL_N_PGROUND, BLEU_PLUS_ONE_ALL_N_AND_BP, BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND};

// Sufficient statistics for computing BLEU; may be for one sentence or a set of sentences.
struct BLEUSuffStats {
	// Total number of n-gram matches, for each order of n (entry k holds number of (k+1)-gram matches).
	vector<int> ngramMatches;
	// Total number of n-grams in translation, for each order of n (entry k holds number of (k+1)-grams).
	vector<int> ngramTotals;
	// Number of tokens in the translation (single sentence or sum of lengths of multiple sentences).
	int translationLength;
	// We also save the reference lengths. This is needed to properly handle multiple references: we
	// only need to keep the length of the reference that is closest to the translation.
	int referenceLength;
};

class MTEval {
	public:
		// multiReferenceLines holds one vector for each reference file and, for each, a vector
		// holding that reference file's lines.
		MTEval(const vector< vector< string > >& multiReferenceLines, Metric metric = STANDARD_BLEU);

		~MTEval();

		// Functions to add reference translations.
		void updateReference(const string& reference);
		void updateReferenceLines(const vector<string>& referenceLines);
		void addReferenceLine(const string& referenceLine);
		// multiReferenceLines holds one vector for each reference file and, for each, a vector
		// holding that reference file's lines.
		void updateMultiReferenceLines(const vector< vector<string> >& multiReferenceLines);

		// Clears internal BLEU sufficient statistics.
		void resetBLEUStats();
		// Adds given BLEU sufficient stats to internal BLEU sufficient statistics.
		void addBLEUStats(const BLEUSuffStats& bleuStats);
		// Subtracts given BLEU sufficient stats from internal BLEU sufficient statistics.
		void subtractBLEUStats(const BLEUSuffStats& bleuStats);
		// Computes corpus-BLEU using current internal BLEU sufficient statistics (e.g., those added using addBLEUStats above).
		double computeBLEU();
		// Computes corpus-BLEU of given translations using stored references added using the constructor.
		double computeBLEU(const vector<string>& translationLines);
		double computeBLEUString(const vector<string>& translationLines, string& printScore);
		// Computes BLEU of the given sentence (translation), placing the computed BLEU sufficient statistics into bleuStats. 
		// This function assumes that all reference translations for the given sentence have already been added to 
		// this MTEval object as the first reference line (i.e., m_multiReferenceLengths and m_referenceLineNGrams 
		// must be filled in and entry 0 is used in each).
		double computeSentenceBLEU(const string& translation, BLEUSuffStats& bleuStats);

	private:
		void initCommon(Metric metric);

		static double computeBrevityPenalty(int translationLength, int referenceLength) {
			double brevityPenalty = 1.0;
			if (translationLength < referenceLength) {
				brevityPenalty = exp(1.0 - (((double) referenceLength) / ((double) translationLength)));
			}
			return brevityPenalty;
		}

		void fillNGramCountsFromTokens(const vector<string>& tokens, map<string, int>& ngramCounts);
		void fillReferenceNGrams(const string& reference, map<string, int>& referenceNGrams, int* refLength);
		void mergeReferenceNGrams(map<string, int>& cumReferenceNGrams, const map<string, int>& newReferenceNGrams);

		//////////// Member variables:

		// BLEU variant to compute.
		Metric m_metric;

		int m_numReferences;
		int m_ngramOrder;

		// Internal BLEU sufficient statistics:
		vector<int> m_ngramMatches;
		vector<int> m_ngramTotals;
		int m_translationLength;
		int m_referenceLength;

		double m_countSmoothValue;

		// Holds n-grams of each line in the reference.
		// If there are multiple references, this holds the maximum count of any n-gram in any of the
		// reference sentences for each line.
		vector< map<string, int> > m_referenceLineNGrams;
		// If there are multiple references, this holds the vector of individual reference sentence lengths
		// for each source sentence.
		// Outer vector has one entry for each line.
		// Inner vector has one entry for each reference file.
		vector< vector<int> > m_multiReferenceLengths;
};

#endif /*MTEVAL_H_*/
