// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.

#include "pbmodel.h"
#include "utils.h"

// This a simpler constructor that can be used for training.
PBModel::PBModel(const string& mosesIniFile, const string& priorMosesIniFile, const string& kbestToIniFileFeatureMap) {
	initCommon();

	bool saveExtraLinesTrue = true;
	bool saveExtraLinesFalse = false;

	// Read the file with mappings from k-best list feature vector feature names to moses.ini feature names.
	m_kBestFeatureTypeToMosesIniFeatureType.clear();
	if (kbestToIniFileFeatureMap == "||NO_FILE_GIVEN||") {
		m_kBestFeatureTypeToMosesIniFeatureType.insert(make_pair("lm", "l"));
		m_kBestFeatureTypeToMosesIniFeatureType.insert(make_pair("tm", "t"));
	} else {
		vector<string> mappings;
		readLinesFromFile(kbestToIniFileFeatureMap, mappings);
		for (unsigned int i = 0; i < mappings.size(); ++i) {
			if (mappings[i].size() > 0) {
				// Tokenize the line around spaces.
				vector<string> mappingTokens;
				tokenize(mappings[i], ' ', mappingTokens);
				if (mappingTokens.size() == 2) {
					m_kBestFeatureTypeToMosesIniFeatureType.insert(make_pair(mappingTokens[0], mappingTokens[1]));
					cerr << "Adding mapping for " << mappingTokens[0] << "-->" << mappingTokens[1] << endl;
				} else {
					cerr << "Error: expected two tokens in feature name mapping (line " << i << " in file "
						 << kbestToIniFileFeatureMap << "). Found " << mappingTokens.size() << " in line: "
						 << endl << mappings[i] << endl;
				}
			}
		}
		if (mappings.size() == 0) {
			cerr << "Error: you specified a mapping file but I couldn't find any valid mappings from it. Specify "
				 << "mappings with \"k-best_feature_name moses_ini_feature_name\", each on a separate line in the "
				 << "mapping file. " << endl << "Using default Moses mappings (lm-->l, tm-->t)." << endl;
			m_kBestFeatureTypeToMosesIniFeatureType.insert(make_pair("lm", "l"));
			m_kBestFeatureTypeToMosesIniFeatureType.insert(make_pair("tm", "t"));
		}
	} 

	// Read the moses.ini file containing initial values of parameter weights.
	if (priorMosesIniFile == "||NO_FILE_GIVEN||") {
		m_priorMeansGiven = false;
	} else {
		m_priorMeansGiven = true;
		readMosesIniFile(priorMosesIniFile, saveExtraLinesFalse);
		m_priorMeans = V;
		m_featureInvertedIndex.clear();
		m_featureIndex.clear();
		V.clear();
		setValue(0, 0.0);
	}

	// Read the moses.ini file.
	readMosesIniFile(mosesIniFile, saveExtraLinesTrue);
}

void PBModel::initCommon() {
	m_negateFeatureWeights = false;
	m_sortTime = 0.0;
	m_totalSentenceTime = 0.0;
	m_numSentencesProcessed = 0;
	m_numTrainingExamplesCompleted = 0;
	V.clear();
	G.clear();
	// Make the first parameter a dummy parameter, with value and gradient component zero.
	setValue(0, 0.0);
	setGradient(0, 0.0);
	m_featureIndex.set_empty_key("DUMMY_INVALID_FEATURE_NAME");
	//m_featureIndex.set_empty_key(NULL);
}

PBModel::~PBModel() {}

string PBModel::getParameterName(PARAM_ID parameterIndex) {
	return m_featureInvertedIndex[parameterIndex];
}

void PBModel::readMosesIniFile(const string& filename, bool saveExtraLines) {
    ifstream mosesIniFile(filename.c_str());
    if (!mosesIniFile) {
   	    cerr << "Cannot open Moses ini file " << filename << ". Exiting..." << endl;
   	    return;
    }

	string str1, str2;
	double featureWeight;

	int nextFreeParamIndex = 1;

	if (saveExtraLines) m_lineNumberToParamIndex.clear();

	string weightPrefix = "[weight-";

	int lineNumber = -1;
	while (!mosesIniFile.eof()) {
		// Get a line from the ini file.
		getline(mosesIniFile, str1);
		if (!mosesIniFile.eof()) {
			// Example file with default moses feature weights:
			/*
			# distortion (reordering) weight
			[weight-d]
			0.3
			0.3
			0.3
			0.3
			0.3
			0.3
			0.3

			# language model weights
			[weight-l]
			0.5000

			# translation model weights
			[weight-t]
			0.2
			0.2
			0.2
			0.2
			0.2

			# word penalty
			[weight-w]
			-1
			*/

			++lineNumber;
			// Save the line from the moses.ini file.
			if (saveExtraLines) m_mosesIniLines.push_back(str1);

			// Get the first 8 characters from str1 to see if they are "[weight-".
			string prefix = str1.substr(0, 8);
			if (prefix == weightPrefix) {
				// Get the feature type, which is the substring starting right after the "-" and 
				// going up to but not including the last instance of "]" in the string.
				unsigned int closeBracketIndex = str1.find_last_of(']');
				bool badLine = false;
				if (closeBracketIndex == string::npos) {
					cerr << "Warning: [weight] line does not contain ']'! Line = " << str1 << endl;
					closeBracketIndex = str1.length() - 1;
					badLine = true;
				}
				string featureType = str1.substr(8, closeBracketIndex - 8);
				if (badLine) {
					cerr << "Possibly buggy config file, but anyway I'm using \"" << featureType << "\" as the feature name." << endl;
				}

				// Get the index of where we have left off in the master parameter vector.
				// This will be the index of the first feature of this type to store in the feature index.
				int featureTypeStartingIndex = nextFreeParamIndex;
				// Below we'll count the number of feature weights provided for this feature type.
				int numFeatures = 0;

				// Get lines until we see a blank line.
				getline(mosesIniFile, str2);
				++lineNumber;
				if (saveExtraLines) m_mosesIniLines.push_back(str2);

				while (str2 != "" && !mosesIniFile.eof()) {
					featureWeight = atof(str2.c_str());
					// Set the feature's weight and initialize the gradient to zero.
					setValue(nextFreeParamIndex, featureWeight);
					setGradient(nextFreeParamIndex, 0.0);
					// Add an entry to the feature inverted index for this feature index and type.
					m_featureInvertedIndex.insert(make_pair(nextFreeParamIndex, featureType));
					if (saveExtraLines) m_lineNumberToParamIndex.insert(make_pair(lineNumber, nextFreeParamIndex));
					++numFeatures;
					++nextFreeParamIndex;

					getline(mosesIniFile, str2);
					++lineNumber;
					if (saveExtraLines) m_mosesIniLines.push_back(str2);
				}
				m_featureIndex.insert(make_pair(featureType, make_pair(featureTypeStartingIndex, numFeatures)));
			}
		}
	}
}

// Writes output moses.ini file.
void PBModel::writeNewMosesIniFile(const string& filename) {
    ofstream mosesIniFile(filename.c_str());
    if (!mosesIniFile.good()) {
   	    cerr << "Cannot open output Moses ini file " << filename << ". Exiting..." << endl;
   	    return;
    }

	//truncateFeatureWeights();
	// First print all extra lines to the file.
	for (unsigned int i = 0; i < m_mosesIniLines.size(); ++i) {
		map<int, int>::iterator iter = m_lineNumberToParamIndex.find(i);
		if (iter != m_lineNumberToParamIndex.end()) {
			// The line was found in the map, so it's a weight line.
			mosesIniFile << getValue(iter->second) << endl;
		} else {
			// The line wasn't found in the map, so it's just an ordinary moses ini file line.
			mosesIniFile << m_mosesIniLines[i] << endl;
		}
	}

	// Go through all feature types.
	/*map<string, pair<int, int> >::const_iterator iter;
	for (iter = m_featureIndex.begin(); iter != m_featureIndex.end(); ++iter) {
		mosesIniFile << "[weight-" << iter->first << "]" << endl;
		unsigned int lastIndex = iter->second.first + iter->second.second;
		for (unsigned int i = iter->second.first; i < lastIndex; ++i) {
			mosesIniFile << getValue(i) << endl;
		}
		mosesIniFile << endl;		
	}*/
}

// Collapse feature weights to zero that would be written with scientific notation.
void PBModel::truncateFeatureWeights() {
	for (PARAM_ID i = 1; i < V.size(); ++i) {
		double weight = getValue(i);
		if (weight < 0.0001 && weight > -0.0001) {
			setValue(i, 0.0);
		}
	}
}

/**
 * Gets parameter value corresponding to given index.
 */
double PBModel::getValue(PARAM_ID index) {
	return V[index];
}

/**
 * Sets parameter value corresponding to given index to given value.
 */
void PBModel::setValue(PARAM_ID index, double newValue) {
	if (index >= V.size()) {
		// Add zeroes to the parameter vector until we reach the desired index.
		while (index > V.size()) {
			V.push_back(0.0);
		}
		// Insert the specified parameter value for the desired index.
		V.push_back(newValue);
	} else {
		V[index] = newValue;
	}
}

/**
 * Gets gradient value corresponding to given index.
 */
double PBModel::getGradient(PARAM_ID index) {
	return G[index];
}

/**
 * Sets gradient value corresponding to given index to given value.
 */
void PBModel::setGradient(PARAM_ID index, double newGradient) {
	if (index >= G.size()) {
		G.push_back(newGradient);
	} else {
		G[index] = newGradient;
	}
}

void PBModel::resetGradientVector(bool logDomain) {
	if (logDomain) {
		for (PARAM_ID i = 0; i < G.size(); i++) {
			G[i] = neg_inf;
		}
	} else {
		for (PARAM_ID i = 0; i < G.size(); i++) {
			G[i] = 0.0;
		}
	}
}

// Subtracts featureMap2 from featureMap1, storing the resulting feature vector in result.
// First clears result.
void PBModel::computeSparseFeatureMapDifference(
		const FEAT_HASH& featureMap1,
		const FEAT_HASH& featureMap2,
		FEAT_HASH& result) {
	result.clear();
	FEAT_HASH::const_iterator iter;
	for (iter = featureMap1.begin(); iter != featureMap1.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] += incrementValue;
	}
	for (iter = featureMap2.begin(); iter != featureMap2.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] -= incrementValue;
	}
}

double PBModel::computeSquaredNormOfSparseFeatureMapDifference(
		const FEAT_HASH& featureMap1,
		const FEAT_HASH& featureMap2) {

	double squaredNorm = 0.0;
	FEAT_HASH result;
	result.set_empty_key(99999);	// TODO: comment out if using sparsehash instead of densehash
	FEAT_HASH::const_iterator iter;
	for (iter = featureMap1.begin(); iter != featureMap1.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] += incrementValue;
	}
	for (iter = featureMap2.begin(); iter != featureMap2.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] -= incrementValue;
	}
	// Go through all entries of result.
	for (iter = result.begin(); iter != result.end(); ++iter) {
		double value = iter->second;
		squaredNorm += (value * value);
	}
	return squaredNorm;
}

void PBModel::printFeatureMapDifference(
		const map<int, double>& featureMap1, const map<int, double>& featureMap2, double stepsize) {

	map<int, double> result;
	for (map<int, double>::const_iterator iter = featureMap1.begin(); iter != featureMap1.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] += incrementValue;
	}
	for (map<int, double>::const_iterator iter = featureMap2.begin(); iter != featureMap2.end(); ++iter) {
		int parameterIndex = iter->first;
		double incrementValue = iter->second;
		result[parameterIndex] -= incrementValue;
	}
	// Go through all entries of result and print them.
	for (map<int, double>::iterator iter = result.begin(); iter != result.end(); ++iter) {
		if (iter->second > 0.00001 || iter->second < -0.00001) {
			cerr << "For parameter " << getParameterName(iter->first) << " update = " << iter->second << " * "
				<< stepsize << " = " << (iter->second * stepsize) << endl;
		}
	}
}

void PBModel::addSparseFeatureValueMapToV(
		const FEAT_HASH& featureMap, double multiplier) {

	FEAT_HASH::const_iterator iter;
	for (iter = featureMap.begin(); iter != featureMap.end(); ++iter) {
		unsigned int parameterIndex = iter->first;
		double incrementValue = iter->second;
		if (parameterIndex >= V.size()) {
			cerr << "Warning: in PBModel::addSparseFeatureValueMapToV(), parameter index " << parameterIndex
				 << " >= V.size() (" << V.size() << ")" << endl;
		} else {
			V[parameterIndex] += (incrementValue * multiplier);
		}
	}
}

void PBModel::subtractSparseFeatureValueMapFromV(
		const FEAT_HASH& featureMap, double multiplier) {

	FEAT_HASH::const_iterator iter;
	for (iter = featureMap.begin(); iter != featureMap.end(); ++iter) {
		unsigned int parameterIndex = iter->first;
		double incrementValue = iter->second;
		if (parameterIndex >= V.size()) {
			cerr << "Warning: in PBModel::subtractSparseFeatureValueMapFromV(), parameter index " << parameterIndex
				 << " >= V.size() (" << V.size() << ")" << endl;
		} else {
			V[parameterIndex] -= (incrementValue * multiplier);
		}
	}
}

void PBModel::loadReferences(const string& refSentsFile, vector<vector<string> >& refSentences) {
	ifstream refSentsFilestream;
	if (fileExists(refSentsFile)) {
		vector<string> currRefLines;
		readLinesFromFile(refSentsFile, currRefLines);
		refSentences.push_back(currRefLines);
	} else {
		int referenceSuffix = 0;
		string currRefSentsFilename = refSentsFile + itos(referenceSuffix);
		while (fileExists(currRefSentsFilename)) {
			vector<string> currRefLines;
			readLinesFromFile(currRefSentsFilename, currRefLines);
			refSentences.push_back(currRefLines);
			++referenceSuffix;
			currRefSentsFilename = refSentsFile + itos(referenceSuffix);
		}
	}
}

