// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.
//
// Implements structured ramp loss minimization (Gimpel & Smith, 2012) for Moses phrase-based and
// hierarchical phrase-based machine translation models given a Moses configuration file and a
// set of k-best lists.

#ifndef PBMODEL_H_
#define PBMODEL_H_

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <set>
#include <stack>
#include <cstdlib>
#include <map>
#include <list>
#include <limits>
#include <math.h>
#include <queue>
#include "mteval.h"
#include "hashing.h"
#include "classias_binary_classifier.h"
#include "../external/include/sparsehash/sparse_hash_map"
#include "../external/include/sparsehash/dense_hash_map"

//#define COMPARE_TRANSLATIONS m_compareLowerMemoryTranslationsObject
//#define FEAT_HASH sparse_hash_map<int, float, IdentityHashUnsignedInt, equnsignedint>
#define FEAT_HASH dense_hash_map<int, float, IdentityHashUnsignedInt, equnsignedint>
//#define FEAT_HASH map<short, float>

using google::sparse_hash_map;
using google::dense_hash_map;

typedef unsigned int PARAM_ID;
typedef string WORD;

enum LossFunction {RAMP1, RAMP2, RAMP3, SOFTRAMP1, SOFTRAMP2, SOFTRAMP3,
	HARDSOFTRAMP, HARDSOFTRAMP_OCH, HARDSOFTRAMP_SMM, BAYESRISK, ALLPAIRS_HINGE,
	ALLPAIRS_SOFTHINGE, EXPLOSS, ALLPAIRS_EXP, ALLPAIRS_EXPCOST, LINEAR_REGRESSION, PRO};
enum OptimizationAlgorithm {SSD, MIRA, PEGASOS};

// To save memory, we only store the translation string, the computed model score, a sparse_hash_map
// holding the feature values, and the computed sentence-level BLEU score.
struct LowMemoryTranslation {
	string text;
	float score;
	FEAT_HASH featureValues;
	float bleuScore;
};

// To save more memory, we forget about the translation string, instead just 
// storing the sentence-level BLEU score in its place. This is all that is 
// needed for most training algorithms.
struct LowerMemoryTranslation {
	float score;
	FEAT_HASH featureValues;
	float bleuScore;
};

struct Translation {
	string text;
	float score;
	map<int, double> featureValues;
	string treeString;
	float bleuScore;
	float corpusBLEUScore;
	BLEUSuffStats bleuSuffStats;
};

struct PROTranslationPair {
	int index1, index2;
	double bleuDifference;
	double absBLEUDifference;
};

// TODO: set this via a compile-time flag.
//typedef LowerMemoryTranslation TRANSLATION;
typedef LowMemoryTranslation TRANSLATION;

// The following struct can be used with sort() to sort Translations in
// decreasing order by score.
struct CompareTranslations {
	bool operator() (const Translation& a, const Translation& b) {
		return a.score > b.score;
	}
};

struct CompareTranslationsByBLEUScore {
	bool operator() (const Translation& a, const Translation& b) {
		return a.bleuScore > b.bleuScore;
	}
};

struct CompareLowMemoryTranslationsByBLEUScore {
	bool operator() (const LowMemoryTranslation& a, const LowMemoryTranslation& b) {
		return a.bleuScore > b.bleuScore;
	}
};

struct CompareLowerMemoryTranslationsByBLEUScore {
	bool operator() (const LowerMemoryTranslation& a, const LowerMemoryTranslation& b) {
		return a.bleuScore > b.bleuScore;
	}
};

// The following struct can be used with sort() to sort PROTranslationPairs in
// decreasing order by score.
struct ComparePROTranslationPairs {
	bool operator() (PROTranslationPair a, PROTranslationPair b) {
		return a.absBLEUDifference > b.absBLEUDifference;
	}
};

class PBModel {
	public:
		PBModel(const string& mosesIniFile, const string& priorMosesIniFile, const string& kbestToIniFileFeatureMap);
		~PBModel();

		// Writes file with feature weights.
		void writeNewMosesIniFile(const string& filename);

		// Does training.
		void train(const string& srcSentsFile, const string& kBestFile, const string& prevKBestFile, const string& refSentsFile,
			const string& outMosesIniFile, unsigned int currIterationNumber, unsigned int numCCCPIters, unsigned int numSubIters,
			const string& lossFunction, const string& costFunctionMetric, const string& optimizationAlgorithm,
			double costCoefficient, double stepsize, double l2Reg, bool fastMode);

	private:
		/**************************************************************************************/
		/********************************** In pbmodel.cpp: ***********************************/
		/**************************************************************************************/
		void initCommon();

		// simple getters and setters for parameter values and their gradients
		double getValue(PARAM_ID index);
		void setValue(PARAM_ID index, double newValue);
		double getGradient(PARAM_ID index);
		void setGradient(PARAM_ID index, double newGradient);
		// Resets G to neg_inf if logDomain, 0.0 otherwise.
		void resetGradientVector(bool logDomain);

		string getParameterName(PARAM_ID parameterIndex);

		// Reads moses.ini file and stores feature weights.
		void readMosesIniFile(const string& filename, bool saveExtraLines);

		void truncateFeatureWeights();

		// Subtracts featureMap2 from featureMap1, storing the resulting feature vector in result. First clears result.
		void computeSparseFeatureMapDifference(
			const FEAT_HASH& featureMap1,
			const FEAT_HASH& featureMap2,
			FEAT_HASH& result);

		double computeSquaredNormOfSparseFeatureMapDifference(
			const FEAT_HASH& featureMap1,
			const FEAT_HASH& featureMap2);

		void printFeatureMapDifference(const map<int, double>& featureMap1, const map<int, double>& featureMap2, double stepsize);

		void addSparseFeatureValueMapToV(
			const FEAT_HASH& featureMap, double multiplier);
		void subtractSparseFeatureValueMapFromV(
			const FEAT_HASH& featureMap, double multiplier);

		void loadReferences(const string& refSentsFile, vector<vector<string> >& refSentences);

		/**************************************************************************************/
		/******************************* In pbmodel_training.cpp: *****************************/
		/**************************************************************************************/

		int readKBestFile(const string& kBestListFile, vector<vector<TRANSLATION> >& kbestLists,
			int numSentences, bool readingPrevKBestLists, bool fastMode, 
			const vector< vector<string> >& refSentences,
			const vector< vector<string> >& refSentencesRotated, 
			int currIterationNumber, const string& extraString);

		void processKBestListFeatureType(vector<string>& tokens, int tokenIndex, int featureIndex, int numFeatures,
			FEAT_HASH& featureValues);

		int findOptimalTranslation(vector<TRANSLATION>& kbestList, bool findMax, bool costAugmented,
			double costMultiplier, bool useModelScoreForMax);

		void rescoreKBestList(vector<TRANSLATION>& kbestList, bool logDifferences);

		void computeBLEUScoresForKBestLists(const vector< vector<string> >& refSentences,
			const vector< vector<string> >& refSentencesRotated, int currIterationNumber, const string& extraString);

		double computeBLEUScoresForKBestListsAfterParameterUpdate(const vector< vector<string> >& refSentences,
			int currIterationNumber, int currCCCPIterationNumber, int subIterationNumber);

		// Sets m_loss to the loss function found in the given string.
		void setLossFunction(const string& lossFunction);

		void copySparseFeatureValues(const FEAT_HASH& source, map<int, double>& dest);

		void sampleAndGenerateTrainingPairsForPRO(const vector<TRANSLATION>& activeKBestList, int currentIndex);

		void proSampleTrainingPairs(const vector<TRANSLATION>& activeKBestList, int gamma,
			double proAlphaThreshold, bool useSigmoid, bool proNoSubsampling, vector<PROTranslationPair>& translationPairs);

		void proAddTrainingPairs(const vector<TRANSLATION>& activeKBestList,
			const vector<PROTranslationPair>& translationPairs, int xi, bool proBothExamples);

		void imputeConcaveTerm(double costCoefficient);

		double computeCost(double costCoefficient, double bleuScore);

		double computeLossGradientForSentence(int i, double costCoefficient);

		int computeSentenceLevelBLEUScores(vector<TRANSLATION>& kBestList, const vector< vector<string> >& currLineMultiReferenceLines);
		int computeSentenceLevelBLEUScores(vector<TRANSLATION>& kBestList, MTEval& mteval, unsigned int kbestListStartingIndex);

		void getReferenceLinesSubset(vector< vector<string> >& full, unsigned int start, unsigned int end, vector< vector<string> >& subset);

		/************************************* Member variables: ****************************************/
		vector<double> V;  // Parameter vector.
		vector<double> G;  // Gradient vector.

		vector<double> m_priorMeans;	// Initial values for feature weights.

		// Maps feature type (part after "weight-" in moses.ini file, e.g., "t", "l", or "d") to a
		// pair holding (1) the start of the feature type in V, and (2) the number of features of this
		// type in V.
		//map<string, pair<int, int> > m_featureIndex;
		dense_hash_map<string, pair<int, int>, MurmurHash2, eqstring> m_featureIndex;

		// Maps feature indexes in V to feature type (see comment above m_featureIndex).
		map<int, string> m_featureInvertedIndex;

		// Maps feature type names in k-best list to weight names in moses ini file.
		map<string, string> m_kBestFeatureTypeToMosesIniFeatureType;

		map<int, int> m_lineNumberToParamIndex;

		bool m_negateFeatureWeights;

		bool m_firstIteration;

		// The following block of variables are member variables used to pass extra
		// arguments to Eval() for use with OWLQN.
		vector<vector<TRANSLATION> > m_kbestLists;
		vector<int> m_imputedIndices;
		LossFunction m_loss;
		Metric m_costMetric;

		ClassiasBinaryClassifier m_binaryClassifier;

		int m_numLBFGSIterations;
		int m_numSubIters;

		int m_numberReferencesReached;
		int m_numberReferencesNotReached;

		int m_referenceLength;
		bool m_priorMeansGiven;

		int m_numTrainingExamplesCompleted;
		vector<string> m_srcSentences, m_tgtSentences;

		vector<set<int> > m_trainingDataParameters;
		int m_currentTrainingExample;

		CompareTranslations m_compareTranslationsObject;
		CompareTranslationsByBLEUScore m_compareTranslationsByBLEUScoreObject;
		CompareLowMemoryTranslationsByBLEUScore m_compareLowMemoryTranslationsObject;
		CompareLowerMemoryTranslationsByBLEUScore m_compareLowerMemoryTranslationsObject;
		ComparePROTranslationPairs m_comparePROTranslationPairsObject;

		/******************** Variables related to files *********************/
		// File names for input files.
		string m_srcSentsFilename;
		string m_srcLatsFilename;
		string m_tgtSentsFilename;
		string m_mosesIniFilename;
		string m_outputModelFilename;

		vector<string> m_mosesIniLines;

		// file handles for input files (read each time a Formula object for a training example is needed)
		ifstream m_srcSentsFile;
		//ifstream m_srcLatsFile;
		ifstream m_tgtSentsFile;
		ifstream m_mosesIniFile;
		//igzstream feature_file;

		/*************** Variables for maintaining cumulative times for various tasks ****************/
		double m_sortTime;

		// The following two variables are for computing the average time to process a
		// sentence for the different algorithms and objectives.
		double m_totalSentenceTime;
		int m_numSentencesProcessed;

};

#endif /*PBMODEL_H_*/
