// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.

#include "rampion.h"

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("rampion -- train a phrase-based or hierarchical phrase-based machine translation model");

    params.addConstraint("src_file", "file containing source sentences. note: only the number of lines is used, so any file "
						 "with the same number of lines as the actual source file will work fine. the number of lines is required "
						 "in situations in which the decoder may not produce translations for certain sentences, or if the translations "
						 "are presented in the n-best list out of order", false); // not optional
    params.addConstraint("kbest_file", "file containing k-best lists of translations of source sentences obtained using weights in in_moses_ini_file", false); // not optional
    params.addConstraint("prev_kbest_file", "file containing k-best lists of translations from all iterations' "
						 "weights *before* the weights in in_moses_ini_file (default: none)", true); // optional
    params.addConstraint("ref_file", "file containing reference translations for source language sentences; "
									 "can be a file stem, with actual reference files denoted ref_file0, ref_file1, etc.", false); // not optional

    params.addConstraint("in_moses_ini_file", "input Moses configuration file", false); // not optional
    params.addConstraint("out_moses_ini_file", "output Moses configuration file with updated weights", false); // not optional
    params.addConstraint("initial_moses_ini_file", "Moses configuration file holding initial feature weights, used for regularization", true); // optional

    params.addConstraint("kbest_to_ini_file_feature_map", "file containing mapping of feature names from those used in "
						 "k-best list feature vectors to those used in Moses ini file; each line of the file should look "
						 "like \"abc def\" where \"abc\" is the name of a feature type in the k-best list feature string "
						 "and \"def\" is the name of the same feature type in the config file (if no file is specified, uses identity "
						 "mappings and the following two used by Moses: \"lm\"->\"l\" and \"tm\"->\"t\")", true); // optional

	params.addConstraint("loss_function", "loss function to minimize; values in "
					     "{pro, bayesrisk, ramp1, ramp2, ramp3, softramp1, softramp2, softramp3, allpairs_hinge, "
						 "allpairs_softhinge, exp, allpairs_exp, allpairs_expcost, linear_regression} (default: ramp3)", true); // optional
	params.addConstraint("curr_train_iter", "current iteration of training, used only for logging purposes (default: 1)", true); // optional

	params.addConstraint("eta", "fixed step size for stochastic subgradient descent (default: 0.0001)", true);
	params.addConstraint("C", "coefficient for L2 regularization (default: 1.0)", true);

	params.addConstraint("num_cccpiters", "number of CCCP outer-loop iterations to run on each training iteration (default: 10)", true); // optional
	params.addConstraint("num_subiters", "number of sub-iterations to run on each CCCP iteration (default: 5)", true); // optional

	params.addConstraint("optimization_algorithm", "algorithm to use for convex optimization within CCCP, applicable for loss functions with multiple "
						 "algorithm choices; values in {ssd, mira, pegasos} (default: ssd)", true); // optional
	params.addConstraint("alpha", "coefficient for cost function (default: 10)", true); // optional
	params.addConstraint("fast_mode", "optimize for speed by computing and printing fewer statistics (default: false)", true); // optional
	params.addConstraint("cost_function_metric", "metric to use in the cost function; choices are:\n"
						 "\t\tSTANDARD_BLEU -- standard unsmoothed BLEU\n"
						 "\t\tLIN_OCH_BLEU_PLUS_ONE -- BLEU+1 from Lin and Och (2004) which only adds one to n-gram counts for n > 1 (used in experiments in NAACL 2012 paper)\n"
						 "\t\tBLEU_PLUS_ONE_ALL_N -- BLEU+1 which adds one for n-grams for all n\n"
						 "\t\tBLEU_PLUS_ONE_ALL_N_PGROUND -- BLEU+1 which adds one for n-grams for all n and also grounds the precision component\n"
						 "\t\tBLEU_PLUS_ONE_ALL_N_AND_BP -- BLEU_PLUS_ONE_ALL_N with adding one to the (selected) reference length\n"
						 "\t\tBLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND -- BLEU_PLUS_ONE_ALL_N_AND_BP with grounding of the precision component\n"
						 "\t\t(default: LIN_OCH_BLEU_PLUS_ONE)", true); // optional

   	if (!params.runConstraints("rampion")) {
       	return 0;
    }

	string srcFile = params.asString("src_file");
	string kbestFile = params.asString("kbest_file");
	string prevKBestFile = params.asString("prev_kbest_file", "||NO_FILE_GIVEN||");
	string refFile = params.asString("ref_file");

	string inMosesIniFile = params.asString("in_moses_ini_file");
	string outMosesIniFile = params.asString("out_moses_ini_file");
	string priorMosesIniFile = params.asString("initial_moses_ini_file", "||NO_FILE_GIVEN||");

	string kbestToIniFileFeatureMap = params.asString("kbest_to_ini_file_feature_map", "||NO_FILE_GIVEN||");

	string lossFunction = params.asString("loss_function", "ramp3");
	int currTrainIter = atoi(params.asString("curr_train_iter", "1").c_str());

	double eta = atof(params.asString("eta", "0.0001").c_str());
	double C = atof(params.asString("C", "1.0").c_str());

	int numCCCPIters = atoi(params.asString("num_cccpiters", "10").c_str());
	int numSubiters = atoi(params.asString("num_subiters", "5").c_str());

	string optimizationAlgorithm = params.asString("optimization_algorithm", "ssd");
	double alpha = atof(params.asString("alpha", "10.0").c_str());

	bool fastMode = (params.asString("fast_mode", "false") == "true");
	string costFunctionMetric = params.asString("cost_function_metric", "LIN_OCH_BLEU_PLUS_ONE");

	struct timeval initialTime, finalTime;
	gettimeofday(&initialTime, 0);

	PBModel pbmodel(inMosesIniFile, priorMosesIniFile, kbestToIniFileFeatureMap);

	// Train the model.
	pbmodel.train(srcFile, kbestFile, prevKBestFile, refFile, outMosesIniFile, currTrainIter,
		numCCCPIters, numSubiters, lossFunction, costFunctionMetric, optimizationAlgorithm, alpha, eta, C, fastMode);

	gettimeofday(&finalTime, 0);
	double totalTime = computeTimeDifference(initialTime, finalTime);
	cerr << "***** Total time: " << totalTime << endl;
    return 0;
}
